#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of python-cw2015 library.
#
# python-cw2015 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-cw2015 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-cw2015.  If not, see <http://www.gnu.org/licenses/>

# 2018-10-01 David Goncalves <david@lestat.st>
#            Python class to communicate with a CW2015 chip on I2C bus.
#            It's a Low-Cost Fuel Gauge IC with Low-SOC Alert from CellWise

import smbus
import sys


class cw2015:

    # Internal register addresses
    REG_VERSION = 0x0
    REG_VCELL = 0x2
    REG_SOC = 0x4
    REG_RRT_ALRT = 0x6
    REG_CONFIG = 0x8
    REG_MODE = 0xa

    # Values for mode register
    MODE_SLEEP = 3 << 6       # Sleep mode
    MODE_QSTRT = 3 << 4       # Quick Start
    MODE_POR = 15             # Power On Reset

    def __init__(self, address=0x62, bus=1, debug=False):
        self._debug = debug
        self.i2c_addr = address
        self.i2c_bus = smbus.SMBus(bus)

        self._debug_message("Accessing CW2015 at address 0x%x on bus %s" % (address, bus))

    def _debug_message(self, message="", error=False):
        """
        Prints debug messages to console when debuging mode is activated
        """

        if (self._debug is True) or (error is True):
            print("%s %s" % (("[DEBUG]" if error is False else "[ERROR]"), message))
            sys.stdout.flush()
            if (error is True):
                raise Exception(message)

    def _write_data(self, address, data):
        """
        Private method to write data to the chip
        """

        self._debug_message("> Writing %s bytes to register 0x%x [%s]" % (len(data), address, ' '.join(["0x%02X" % x for x in data])))
        self.i2c_bus.write_i2c_block_data(self.i2c_addr, address, data)

    def _read_data(self, address, size):
        """
        Private method to read data from the chip
        """

        self._debug_message("> Reading %s bytes from register 0x%x" % (size, address))
        data = self.i2c_bus.read_i2c_block_data(self.i2c_addr, address, size)
        self._debug_message("> CW2015 response : [%s]" % ' '.join(["0x%02X" % x for x in data]))
        return(data)

    def GetFirmwareVersion(self):
        """
        Returns the chip's firmware version
        """

        self._debug_message("Reading firmware version")
        data = self._read_data(self.REG_VERSION, 1)
        self._debug_message("CW2015 version : %d" % data[0])
        return(data[0])

    def GetCellVoltage(self):
        """
        Returns the cell voltage in millivolts
        """

        self._debug_message("Reading cell voltage")
        data = self._read_data(self.REG_VCELL, 2)
        voltage = ((data[1] & 0x3F) + (data[0] << 8)) * 0.305
        self._debug_message("Cell voltage is %smV" % voltage)
        return(int(voltage))

    def GetStateOfCharge(self):
        """
        Get the SOC by reading the appropriate register.

        This method uses the high precision value from the
        register and round it to two decimals.
        """

        self._debug_message("Reading State Of Charge")
        data = self._read_data(self.REG_SOC, 2)
        soc = round(data[0] + (data[1] * (1/256.0)), 2)
        self._debug_message("SOC = %s%%" % soc)
        return(soc)

    def GetRemainingRunTime(self):
        """
        Returns the Remaining RunTime (RRT) value calculated by the IC
        using minutes as units.

        The result is a dict containing 'runtime' for the time remaining and
        an 'alert' bool that indicates if the ALRT bit is triggered.
        """

        self._debug_message("Reading Remaining Runtime")
        data = self._read_data(self.REG_RRT_ALRT, 2)

        alert_bit = data[1] >> 7
        runtime = ((data[0] & 0x1F) << 8) + data[1]

        self._debug_message("Estimated Runtime is %s minutes (ALERT flag is %s)" % (runtime, ("on" if alert_bit is True else "off")))
        return({"runtime": runtime, "alert": bool(alert_bit)})

    def GetAlertThreshold(self):
        """
        Returns the Alert Threshold value set on the config register.
        This threshold is configurable from 0% to 31% and activates the
        alert bit when SoC is below this value (default: 3%)
        """

        self._debug_message("Reading alert threshold from config register")
        data = self._read_data(self.REG_CONFIG, 1)
        threshold = data[0] >> 3
        self._debug_message("Threshold set to %s%%" % threshold)
        return(threshold)

    def SetAlertThreshold(self, threshold=3):
        """
        Set the Alert Threshold value on config register (0->31%)
        """

        self._debug_message("Setting alert threshold to %s%%" % threshold)
        if threshold not in range(0, 1 << 5):
            raise ValueError("Threshold must be between 0 and 31")

        self._write_data(self.REG_CONFIG, [threshold << 3])

    def SetModeRegister(self, mode_mask=0):
        """
        Set options to the mode register.
        This method will read the current value and 'add' the desired flag
        """

        self._debug_message("Configuring mode register with '0x%x'" % mode_mask)
        data = self._read_data(self.REG_MODE, 1)
        self._debug_message("Current mode value is '0x%x'" % data[0])
        self._write_data(self.REG_MODE, [data[0] | mode_mask])
        self._debug_message("Setting mode to '0x%x'" % (data[0] | mode_mask))
